"""The Simple Management Protocol (SMP) OS Management group."""


from enum import IntEnum, auto, unique
from typing import Any, Dict

from pydantic import BaseModel, ConfigDict, Field

from smp import error, header, message


class _OSManagementGroup:
    _GROUP_ID = header.GroupId.OS_MANAGEMENT


class EchoWriteRequest(_OSManagementGroup, message.WriteRequest):
    _COMMAND_ID = header.CommandId.OSManagement.ECHO

    d: str


class EchoWriteResponse(_OSManagementGroup, message.WriteResponse):
    _COMMAND_ID = header.CommandId.OSManagement.ECHO

    r: str


class ResetWriteRequest(_OSManagementGroup, message.WriteRequest):
    _COMMAND_ID = header.CommandId.OSManagement.RESET


class ResetWriteResponse(_OSManagementGroup, message.WriteResponse):
    _COMMAND_ID = header.CommandId.OSManagement.RESET


class TaskStatisticsReadRequest(_OSManagementGroup, message.ReadRequest):
    _COMMAND_ID = header.CommandId.OSManagement.TASK_STATS


class TaskStatistics(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)

    prio: int
    """Task priority."""
    tid: int
    """Numeric task ID."""
    state: int
    """Numeric task state."""
    stkuse: int
    """Stack usage.

    The unit is system dependent and in case of Zephyr this is number of 4 byte words.
    """
    stksize: int
    """Stack size.

    The unit is system dependent and in case of Zephyr this is number of 4 byte words.
    """
    cswcnt: int
    """Number of context switches."""
    runtime: int
    """Runtime in ticks."""
    last_checkin: int
    """Set to 0 by Zephyr."""
    next_checkin: int
    """Set to 0 by Zephyr."""


class TaskStatisticsReadResponse(_OSManagementGroup, message.ReadResponse):
    _COMMAND_ID = header.CommandId.OSManagement.TASK_STATS

    tasks: Dict[str, TaskStatistics]


class MemoryPoolStatisticsReadRequest(_OSManagementGroup, message.ReadRequest):
    _COMMAND_ID = header.CommandId.OSManagement.MEMORY_POOL_STATS


class MemoryPoolStatistics(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)

    blksize: int
    """Size of the memory block in the pool."""
    nblks: int
    """Number of memory blocks in the pool."""
    nfree: int
    """Number of free memory blocks in the pool."""
    min: int
    """Lowest number of free blocks the pool reached during run-time."""


class MemoryPoolStatisticsReadResponse(_OSManagementGroup, message.ReadResponse):
    """The memory pools are accessed by name."""

    model_config = ConfigDict(extra="allow", frozen=True)

    _COMMAND_ID = header.CommandId.OSManagement.MEMORY_POOL_STATS


class DateTimeReadRequest(_OSManagementGroup, message.ReadRequest):
    _COMMAND_ID = header.CommandId.OSManagement.DATETIME_STRING


class DateTimeReadResponse(_OSManagementGroup, message.ReadResponse):
    _COMMAND_ID = header.CommandId.OSManagement.DATETIME_STRING

    datetime: str


class DateTimeWriteRequest(_OSManagementGroup, message.WriteRequest):
    _COMMAND_ID = header.CommandId.OSManagement.DATETIME_STRING

    datetime: str


class DateTimeWriteResponse(_OSManagementGroup, message.WriteResponse):
    _COMMAND_ID = header.CommandId.OSManagement.DATETIME_STRING


class MCUMgrParametersReadRequest(_OSManagementGroup, message.ReadRequest):
    _COMMAND_ID = header.CommandId.OSManagement.MCUMGR_PARAMETERS


class MCUMgrParametersReadResponse(_OSManagementGroup, message.ReadResponse):
    _COMMAND_ID = header.CommandId.OSManagement.MCUMGR_PARAMETERS

    buf_size: int
    buf_count: int


class OSApplicationInfoReadRequest(_OSManagementGroup, message.ReadRequest):
    _COMMAND_ID = header.CommandId.OSManagement.OS_APPLICATION_INFO

    format: str | None = None
    """Format specifier of returned response.

    Fields are appended in their natural ascending index order, not the order
    of characters that are received by the command.

    Format specifiers:
    * `s` Kernel name
    * `n` Node name
    * `r` Kernel release
    * `v` Kernel version
    * `b` Build date and time (requires `CONFIG_MCUMGR_GRP_OS_INFO_BUILD_DATE_TIME`)
    * `m` Machine
    * `p` Processor
    * `i` Hardware platform
    * `o` Operating system
    * `a` All fields (shorthand for all above options)

    If this option is not provided, the `s` Kernel name option will be used.
    """


class OSApplicationInfoReadResponse(_OSManagementGroup, message.ReadResponse):
    _COMMAND_ID = header.CommandId.OSManagement.OS_APPLICATION_INFO

    output: str


class BootloaderInformationReadRequest(_OSManagementGroup, message.ReadRequest):
    _COMMAND_ID = header.CommandId.OSManagement.BOOTLOADER_INFO

    query: str | None = None
    """Is string representing query for parameters.

    With no restrictions how the query looks like as processing of query is left
    for bootloader backend. If there is no query, then response will return string
    identifying the bootloader.

    MCUboot supports the query string,"mode".  The response to mode is of type
    `MCUbootMode`.
    """


@unique
class MCUbootMode(IntEnum):
    UNKNOWN = -1
    APPLICATION = 0
    SWAP_USING_SCRATCH = 1
    OVERWRITE_ONLY = 2
    SWAP_WITHOUT_SCRATCH = 3
    DIRECT_XIP_WITHOUT_REVERT = 4
    DIRECT_XIP_WITH_REVERT = 5
    RAM_LOADER = 6


class MCUbootModeQueryResponse(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)

    mode: MCUbootMode
    no_downgrade: bool | None = Field(alias="no-downgrade", default=None)


class BootloaderInformationReadResponse(_OSManagementGroup, message.ReadResponse):
    _COMMAND_ID = header.CommandId.OSManagement.BOOTLOADER_INFO

    bootloader: str
    """String identifying the bootloader.  For MCUboot it will be "MCUboot"."""
    response: MCUbootModeQueryResponse | Any | None = None
    """Response to “query”.

    This is optional and may be left out in case when query yields no response,
    SMP version 2 error code of `OS_MGMT_ERR_QUERY_YIELDS_NO_ANSWER` is
    expected. Response may have more than one parameter reported back or it may
    be a map, that is dependent on bootloader backend and query."""


@unique
class OS_MGMT_RET_RC(IntEnum):
    OK = 0
    """No error, this is implied if there is no ret value in the response."""

    UNKNOWN = auto()
    """Unknown error occurred."""

    INVALID_FORMAT = auto()
    """The provided format value is not valid."""

    QUERY_YIELDS_NO_ANSWER = auto()
    """Query was not recognized."""

    RTC_NOT_SET = auto()
    """RTC is not set."""

    RTC_COMMAND_FAILED = auto()
    """RTC command failed."""


class OSManagementErrorV0(error.ErrorV0[OS_MGMT_RET_RC]):
    _GROUP_ID = header.GroupId.OS_MANAGEMENT


class OSManagementErrorV1(error.ErrorV1[OS_MGMT_RET_RC]):
    _GROUP_ID = header.GroupId.OS_MANAGEMENT



"""The Simple Management Protocol (SMP) Message base class."""

import itertools
from abc import ABC
from enum import IntEnum, unique
from typing import ClassVar, Type, TypeVar, cast

import cbor2
from pydantic import BaseModel, ConfigDict

from smp import header as smpheader
from smp.exceptions import SMPMismatchedGroupId

T = TypeVar("T", bound='_MessageBase')


_counter = itertools.count()


class _MessageBase(ABC, BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)

    _OP: ClassVar[smpheader.OP]
    _VERSION: ClassVar[smpheader.Version] = smpheader.Version.V0
    _FLAGS: ClassVar[smpheader.Flag] = smpheader.Flag(0)
    _GROUP_ID: ClassVar[smpheader.GroupId]
    _COMMAND_ID: ClassVar[
        smpheader.CommandId.ImageManagement
        | smpheader.CommandId.OSManagement
        | smpheader.CommandId.ShellManagement
        | smpheader.CommandId.Intercreate
    ]

    header: smpheader.Header | None = None

    def __bytes__(self) -> bytes:
        return self._bytes  # type: ignore

    @property
    def BYTES(self) -> bytes:
        return self._bytes  # type: ignore

    @classmethod
    def loads(cls: Type[T], data: bytes) -> T:
        """Deserialize the SMP message."""
        message = cls(
            header=smpheader.Header.loads(data[: smpheader.Header.SIZE]),
            **cast(dict, cbor2.loads(data[smpheader.Header.SIZE :])),
        )
        if message.header is None:  # pragma: no cover
            raise ValueError
        if message.header.group_id != cls._GROUP_ID:  # pragma: no cover
            raise SMPMismatchedGroupId(
                f"{cls.__name__} has {cls._GROUP_ID}, header has {message.header.group_id}"
            )
        return message

    @classmethod
    def load(cls: Type[T], header: smpheader.Header, data: dict) -> T:
        """Load an SMP header and CBOR dict."""
        if header.group_id != cls._GROUP_ID:  # pragma: no cover
            raise SMPMismatchedGroupId(
                f"{cls.__name__} has {cls._GROUP_ID}, header has {header.group_id}"
            )
        return cls(header=header, **data)


class Request(_MessageBase, ABC):
    def model_post_init(self, _: None) -> None:
        data_bytes = cbor2.dumps(
            self.model_dump(exclude_unset=True, exclude={'header'}, exclude_none=True)
        )
        if self.header is None:
            object.__setattr__(
                self,
                'header',
                smpheader.Header(
                    op=self._OP,
                    version=self._VERSION,
                    flags=smpheader.Flag(self._FLAGS),
                    length=len(data_bytes),
                    group_id=smpheader.GroupId(self._GROUP_ID),
                    sequence=next(_counter) % 0xFF,
                    command_id=smpheader.Header._MAP_GROUP_ID_TO_COMMAND_ID_ENUM[
                        smpheader.GroupId(self._GROUP_ID)
                    ](self._COMMAND_ID),
                ),
            )
        self._bytes = cast(smpheader.Header, self.header).BYTES + data_bytes


@unique
class ResponseType(IntEnum):
    """An SMP `Response` to an SMP `Request` must be `SUCCESS`, `ERROR_V0`, or `ERROR_V1`."""

    SUCCESS = 0
    ERROR_V0 = 1
    ERROR_V1 = 2


class Response(_MessageBase, ABC):
    sequence: int = 0

    RESPONSE_TYPE: ClassVar[ResponseType]

    def model_post_init(self, _: None) -> None:
        data_bytes = cbor2.dumps(
            self.model_dump(exclude_unset=True, exclude={'header', 'sequence'}, exclude_none=True)
        )
        if self.header is None:
            object.__setattr__(
                self,
                'header',
                smpheader.Header(
                    op=self._OP,
                    version=self._VERSION,
                    flags=smpheader.Flag(self._FLAGS),
                    length=len(data_bytes),
                    group_id=smpheader.GroupId(self._GROUP_ID),
                    sequence=self.sequence,
                    command_id=smpheader.Header._MAP_GROUP_ID_TO_COMMAND_ID_ENUM[
                        smpheader.GroupId(self._GROUP_ID)
                    ](self._COMMAND_ID),
                ),
            )
        self._bytes = cast(smpheader.Header, self.header).BYTES + data_bytes


class ReadRequest(Request, ABC):
    """A read request from an SMP client to an SMP server."""

    _OP = smpheader.OP.READ


class ReadResponse(Response, ABC):
    """A response from an SMP server to an SMP client read request."""

    RESPONSE_TYPE = ResponseType.SUCCESS
    _OP = smpheader.OP.READ_RSP


class WriteRequest(Request, ABC):
    """A write request from an SMP client to an SMP server."""

    _OP = smpheader.OP.WRITE


class WriteResponse(Response, ABC):
    """A response from an SMP server to an SMP client write request."""

    RESPONSE_TYPE = ResponseType.SUCCESS
    _OP = smpheader.OP.WRITE_RSP



"""The Simple Management Protocol (SMP) Shell Management group."""


from enum import IntEnum, auto, unique
from typing import List

from smp import error, header, message


class _ShellManagementGroup:
    _GROUP_ID = header.GroupId.SHELL_MANAGEMENT


class ExecuteRequest(_ShellManagementGroup, message.WriteRequest):
    _COMMAND_ID = header.CommandId.ShellManagement.EXECUTE

    argv: List[str]


class ExecuteResponse(_ShellManagementGroup, message.WriteResponse):
    _COMMAND_ID = header.CommandId.ShellManagement.EXECUTE

    o: str
    ret: int


@unique
class SHELL_MGMT_RET_RC(IntEnum):
    OK = 0
    """No error, this is implied if there is no ret value in the response."""

    UNKNOWN = auto()
    """Unknown error occurred."""

    INVALID_FORMAT = auto()
    """The provided format value is not valid."""


class ShellManagementErrorV0(error.ErrorV0[SHELL_MGMT_RET_RC]):
    _GROUP_ID = header.GroupId.SHELL_MANAGEMENT


class ShellManagementErrorV1(error.ErrorV1[SHELL_MGMT_RET_RC]):
    _GROUP_ID = header.GroupId.SHELL_MANAGEMENT



"""The Simple Management Protocol (SMP) error responses."""

from enum import IntEnum
from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict

from smp import message
from smp.header import GroupId

T = TypeVar("T", bound=IntEnum)


class ErrorV0(message.Response, Generic[T]):
    RESPONSE_TYPE = message.ResponseType.ERROR_V0

    rc: T
    rsn: str | None = None


class Err(BaseModel, Generic[T]):
    model_config = ConfigDict(extra="forbid", frozen=True, arbitrary_types_allowed=True)

    group: GroupId
    rc: T


class ErrorV1(message.Response, Generic[T]):
    RESPONSE_TYPE = message.ResponseType.ERROR_V1

    err: Err[T]



"""The Simple Management Protocol (SMP) Intercreate Management group."""

from enum import IntEnum, auto, unique

from smp import error, header, message


class _IntercreateManagementGroup:
    _GROUP_ID = header.GroupId.INTERCREATE


class ImageUploadWriteRequest(_IntercreateManagementGroup, message.WriteRequest):
    _COMMAND_ID = header.CommandId.Intercreate.UPLOAD

    off: int
    """The offset in the image to write to."""

    data: bytes
    """The data to write to the image."""

    image: int | None = None
    """The image to write to; required when off == 0."""

    len: int | None = None
    """The length of the data to write; required when off == 0."""

    sha: bytes | None = None
    """The SHA-256 hash of the image; optional when off == 0, else ignored."""


class ImageUploadWriteResponse(_IntercreateManagementGroup, message.WriteResponse):
    _COMMAND_ID = header.CommandId.Intercreate.UPLOAD

    off: int
    """The offset in the image after the request was written."""


@unique
class IC_MGMT_ERR(IntEnum):
    OK = 0
    """No error."""

    INVALID_IMAGE = auto()
    """No image matched the image provided."""


class ImageManagementErrorV0(error.ErrorV0[IC_MGMT_ERR]):
    _GROUP_ID = header.GroupId.INTERCREATE


class ImageManagementErrorV1(error.ErrorV1[IC_MGMT_ERR]):
    _GROUP_ID = header.GroupId.INTERCREATE



"""The Simple Management Protocol (SMP) header."""

import struct
from dataclasses import dataclass
from enum import IntEnum, IntFlag, auto, unique
from typing import TypeVar


class CommandId:
    @unique
    class OSManagement(IntEnum):
        ECHO = 0
        ECHO_CONTROL = 1
        TASK_STATS = 2
        MEMORY_POOL_STATS = 3
        DATETIME_STRING = 4
        RESET = 5
        MCUMGR_PARAMETERS = 6
        OS_APPLICATION_INFO = 7
        BOOTLOADER_INFO = 8

    @unique
    class ImageManagement(IntEnum):
        STATE = 0
        UPLOAD = 1
        FILE = 2
        CORELIST = 3
        CORELOAD = 4
        ERASE = 5

    @unique
    class ShellManagement(IntEnum):
        EXECUTE = 0

    @unique
    class Intercreate(IntEnum):
        UPLOAD = 1


AnyCommandId = TypeVar("AnyCommandId", bound=IntEnum)


class GroupId(IntEnum):
    OS_MANAGEMENT = 0
    IMAGE_MANAGEMENT = 1
    STATISTICS_MANAGEMENT = 2
    SETTINGS_MANAGEMENT = 3
    LOG_MANAGEMENT = 4
    RUNTIME_TESTS = 5
    SPLIT_IMAGE_MANAGEMENT = 6
    TEST_CRASH = 7
    FILE_MANAGEMENT = 8
    SHELL_MANAGEMENT = 9
    ZEPHYR = 63
    _APPLICATIION_CUSTOM_MIN = 64
    INTERCREATE = 64


@unique
class OP(IntEnum):
    READ = 0
    READ_RSP = 1
    WRITE = 2
    WRITE_RSP = 3


class _OP_BIT:
    MASK = 0b111
    SHIFT = 0


@unique
class Version(IntEnum):
    V0 = 0
    V1 = 1


class _VERSION_BIT:
    MASK = 0b11
    SHIFT = 3


@unique
class Flag(IntFlag):
    UNUSED = auto()


@dataclass(frozen=True)
class Header:
    op: OP
    version: Version
    flags: Flag
    length: int
    group_id: GroupId
    sequence: int
    command_id: (
        CommandId.OSManagement | CommandId.ImageManagement | CommandId.ShellManagement | IntEnum
    )

    _MAP_GROUP_ID_TO_COMMAND_ID_ENUM = {
        GroupId.OS_MANAGEMENT: CommandId.OSManagement,
        GroupId.IMAGE_MANAGEMENT: CommandId.ImageManagement,
        GroupId.SHELL_MANAGEMENT: CommandId.ShellManagement,
        GroupId.INTERCREATE: CommandId.Intercreate,
    }
    _STRUCT = struct.Struct("!BBHHBB")
    SIZE = _STRUCT.size

    @staticmethod
    def _pack_op(op: OP) -> int:
        """The value to be packed into the byte."""
        return op << _OP_BIT.SHIFT

    @staticmethod
    def _unpack_op(res_ver_op_byte: int) -> OP:
        """The value unpacked from the byte."""
        return OP((res_ver_op_byte & _OP_BIT.MASK) >> _OP_BIT.SHIFT)

    @staticmethod
    def _pack_version(version: Version) -> int:
        """The value to be packed into the byte."""
        return version << _VERSION_BIT.SHIFT

    @staticmethod
    def _unpack_version(res_ver_op_byte: int) -> Version:
        """The value unpacked from the byte."""
        return Version((res_ver_op_byte >> _VERSION_BIT.SHIFT) & _VERSION_BIT.MASK)

    @staticmethod
    def _pack_op_and_version(op: OP, version: Version) -> int:
        """The op and version packed into one byte."""
        return Header._pack_op(op) | Header._pack_version(version)

    def __post_init__(self) -> None:
        object.__setattr__(
            self,
            '_bytes',
            self._STRUCT.pack(
                self._pack_op_and_version(self.op, self.version),
                Flag(self.flags),
                self.length,
                GroupId(self.group_id),
                self.sequence,
                Header._MAP_GROUP_ID_TO_COMMAND_ID_ENUM[GroupId(self.group_id)](self.command_id),
            ),
        )

    def __bytes__(self) -> bytes:
        return self._bytes  # type: ignore

    @property
    def BYTES(self) -> bytes:
        return self._bytes  # type: ignore

    @staticmethod
    def loads(header: bytes) -> 'Header':
        """Deserialize the header bytes to a `Header`."""
        assert len(header) == 8, "The header is specified as 8 bytes"

        res_ver_op_byte, flags, length, group_id, sequence, command_id = Header._STRUCT.unpack(
            header
        )

        return Header(
            Header._unpack_op(res_ver_op_byte),
            Header._unpack_version(res_ver_op_byte),
            Flag(flags),
            length,
            GroupId(group_id),
            sequence,
            Header._MAP_GROUP_ID_TO_COMMAND_ID_ENUM[GroupId(group_id)](command_id),
        )



"""The Simple Management Protocol (SMP) Image Management group."""

from enum import IntEnum, auto, unique
from typing import Generator, List

from pydantic import BaseModel, ConfigDict, ValidationInfo, field_validator

from smp import error, header, message


class _ImageManagementGroup:
    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT


class HashBytes(bytes):  # pragma: no cover
    """Only to print something useful to the console."""

    def __rich_repr__(self) -> Generator[str, None, None]:
        yield self.hex().upper()


class ImageState(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True, arbitrary_types_allowed=True)

    slot: int
    version: str
    image: int | None = None
    hash: HashBytes | bytes | None = None
    bootable: bool | None = None
    pending: bool | None = None
    confirmed: bool | None = None
    active: bool | None = None
    permanent: bool | None = None

    @field_validator("hash")
    @classmethod
    def cast_bytes(cls, v: bytes | None, _: ValidationInfo) -> HashBytes | None:
        if v is None:
            return None
        return HashBytes(v)


class ImageStatesReadRequest(_ImageManagementGroup, message.ReadRequest):
    _COMMAND_ID = header.CommandId.ImageManagement.STATE


class ImageStatesReadResponse(_ImageManagementGroup, message.ReadResponse):
    _COMMAND_ID = header.CommandId.ImageManagement.STATE

    images: List[ImageState]
    splitStatus: int | None = None


class ImageStatesWriteRequest(_ImageManagementGroup, message.WriteRequest):
    _COMMAND_ID = header.CommandId.ImageManagement.STATE

    hash: bytes
    confirm: bool = False


class ImageStatesWriteResponse(ImageStatesReadResponse):
    pass


class ImageUploadWriteRequest(_ImageManagementGroup, message.WriteRequest):
    _COMMAND_ID = header.CommandId.ImageManagement.UPLOAD

    off: int
    data: bytes
    image: int | None = None  # 'required' when off == 0
    len: int | None = None  # required when off == 0
    sha: bytes | None = None  # should be sent when off == 0
    upgrade: bool | None = None  # allowed when off == 0


class ImageUploadProgressWriteResponse(_ImageManagementGroup, message.WriteResponse):
    _COMMAND_ID = header.CommandId.ImageManagement.UPLOAD

    rc: int | None = None
    off: int | None = None


class ImageUploadFinalWriteResponse(_ImageManagementGroup, message.WriteResponse):
    _COMMAND_ID = header.CommandId.ImageManagement.UPLOAD

    off: int | None = None
    match: bool | None = None


class ImageEraseRequest(_ImageManagementGroup, message.WriteRequest):
    _COMMAND_ID = header.CommandId.ImageManagement.ERASE

    slot: int | None = None
    """The slot to erase. If not provided, slot 1 will be erased."""


class ImageEraseResponse(_ImageManagementGroup, message.WriteResponse):
    _COMMAND_ID = header.CommandId.ImageManagement.ERASE


@unique
class IMG_MGMT_ERR(IntEnum):
    OK = 0
    """No error, this is implied if there is no ret value in the response"""

    UNKNOWN = auto()
    """Unknown error occurred."""

    FLASH_CONFIG_QUERY_FAIL = auto()
    """Failed to query flash area configuration."""

    NO_IMAGE = auto()
    """There is no image in the slot."""

    NO_TLVS = auto()
    """The image in the slot has no TLVs (tag, length, value)."""

    INVALID_TLV = auto()
    """The image in the slot has an invalid TLV type and/or length."""

    TLV_MULTIPLE_HASHES_FOUND = auto()
    """The image in the slot has multiple hash TLVs, which is invalid."""

    TLV_INVALID_SIZE = auto()
    """The image in the slot has an invalid TLV size."""

    HASH_NOT_FOUND = auto()
    """The image in the slot does not have a hash TLV, which is required."""

    NO_FREE_SLOT = auto()
    """There is no free slot to place the image."""

    FLASH_OPEN_FAILED = auto()
    """Flash area opening failed."""

    FLASH_READ_FAILED = auto()
    """Flash area reading failed."""

    FLASH_WRITE_FAILED = auto()
    """Flash area writing failed."""

    FLASH_ERASE_FAILED = auto()
    """Flash area erase failed."""

    INVALID_SLOT = auto()
    """The provided slot is not valid."""

    NO_FREE_MEMORY = auto()
    """Insufficient heap memory (malloc failed)."""

    FLASH_CONTEXT_ALREADY_SET = auto()
    """The flash context is already set."""

    FLASH_CONTEXT_NOT_SET = auto()
    """The flash context is not set."""

    FLASH_AREA_DEVICE_NULL = auto()
    """The device for the flash area is NULL."""

    INVALID_PAGE_OFFSET = auto()
    """The offset for a page number is invalid."""

    INVALID_OFFSET = auto()
    """The offset parameter was not provided and is required."""

    INVALID_LENGTH = auto()
    """The length parameter was not provided and is required."""

    INVALID_IMAGE_HEADER = auto()
    """The image length is smaller than the size of an image header."""

    INVALID_IMAGE_HEADER_MAGIC = auto()
    """The image header magic value does not match the expected value."""

    INVALID_HASH = auto()
    """The hash parameter provided is not valid."""

    INVALID_FLASH_ADDRESS = auto()
    """The image load address does not match the address of the flash area."""

    VERSION_GET_FAILED = auto()
    """Failed to get version of currently running application."""

    CURRENT_VERSION_IS_NEWER = auto()
    """The currently running application is newer than the version being uploaded."""

    IMAGE_ALREADY_PENDING = auto()
    """There is already an image operating pending."""

    INVALID_IMAGE_VECTOR_TABLE = auto()
    """The image vector table is invalid."""

    INVALID_IMAGE_TOO_LARGE = auto()
    """The image it too large to fit."""

    INVALID_IMAGE_DATA_OVERRUN = auto()
    """The amount of data sent is larger than the provided image size."""

    IMAGE_CONFIRMATION_DENIED = auto()
    """Confirmation of image has been denied"""

    IMAGE_SETTING_TEST_TO_ACTIVE_DENIED = auto()
    """Setting test to active slot is not allowed"""


class ImageManagementErrorV0(error.ErrorV0[IMG_MGMT_ERR]):
    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT


class ImageManagementErrorV1(error.ErrorV1[IMG_MGMT_ERR]):
    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT



"""Simple Management Protocol (SMP) runtime Python `Exception`s."""


class SMPException(Exception):
    ...


class SMPDecodeError(SMPException):
    ...


class SMPBadStartDelimiter(SMPDecodeError):
    ...


class SMPBadContinueDelimiter(SMPDecodeError):
    ...


class SMPBadCRC(SMPDecodeError):
    ...


class SMPDeserializationError(SMPException):
    ...


class SMPMismatchedGroupId(SMPDeserializationError):
    ...



"""Generators for encoding and decoding Simple Management Protocol (SMP) packets."""

import logging
import struct
from base64 import b64decode, b64encode
from typing import Final, Generator

from crcmod.predefined import mkPredefinedCrcFun  # type: ignore

from smp.exceptions import SMPBadContinueDelimiter, SMPBadCRC, SMPBadStartDelimiter

SIXTY_NINE: Final = struct.pack("!BB", 6, 9)
FOUR_TWENTY: Final = struct.pack("!BB", 4, 20)
DELIMITER_SIZE: Final = len(SIXTY_NINE)
assert len(SIXTY_NINE) == len(FOUR_TWENTY) and len(FOUR_TWENTY) == DELIMITER_SIZE
CR: Final = b"\n"
LINE_LENGTH_SUBTRACTOR: Final = 4  # you'd think it's 3, len(delimiter) + len(CR)
FRAME_LENGTH_STRUCT: Final = struct.Struct("!H")
CRC16_STRUCT: Final = struct.Struct("!H")

crc16_func: Final = mkPredefinedCrcFun("xmodem")


logger: Final = logging.getLogger(__name__)


def encode(request: bytes, line_length: int = 8192) -> Generator[bytes, None, None]:
    """Iteratively pack an SMP bytes to packets of `line_length` size."""

    logger.debug(f"Serializing {request=}")

    crc16 = CRC16_STRUCT.pack(crc16_func(request))
    frame_length = FRAME_LENGTH_STRUCT.pack(len(request) + CRC16_STRUCT.size)

    total_size = FRAME_LENGTH_STRUCT.size + len(request) + CRC16_STRUCT.size
    logger.debug(f"Total size of request is {total_size}B")

    packet_size = ((line_length - LINE_LENGTH_SUBTRACTOR) // 4) * 4

    complete_b64 = b64encode(frame_length + request + crc16)
    logger.debug(f"Total size of b64 encoded request is {len(complete_b64)}B")

    # send the start delimiter, as many bytes as possible, and newline
    yield SIXTY_NINE + complete_b64[:packet_size] + CR

    remaining = complete_b64[packet_size:]

    while len(remaining) > 0:
        # send the continue delimiter, as many bytes as possible, and newline
        yield FOUR_TWENTY + remaining[:packet_size] + CR
        remaining = remaining[packet_size:]


def decode() -> Generator[None, bytes, bytes]:
    """Iteratively unpack a series of SMP packets to bytes."""

    packet = yield

    if packet[:DELIMITER_SIZE] != SIXTY_NINE:
        raise SMPBadStartDelimiter(f"Bad start delimiter {packet[:DELIMITER_SIZE].hex()}")

    length_and_frame = b64decode(packet[DELIMITER_SIZE : -len(CR)])
    frame_length = FRAME_LENGTH_STRUCT.unpack(length_and_frame[: FRAME_LENGTH_STRUCT.size])[0]
    logger.debug(f"Response {length_and_frame=}")

    frame = bytearray(length_and_frame[FRAME_LENGTH_STRUCT.size :])
    logger.debug(f"length of frame: {len(frame)}")
    while len(frame) < frame_length:
        packet = yield

        if packet[:DELIMITER_SIZE] != FOUR_TWENTY:
            raise SMPBadContinueDelimiter(f"Bad continue delimiter {packet[:DELIMITER_SIZE].hex()}")

        frame.extend(b64decode(packet[DELIMITER_SIZE : -len(CR)]))

    complete_frame = frame[: -CRC16_STRUCT.size]
    packet_crc16 = CRC16_STRUCT.unpack(frame[-CRC16_STRUCT.size :])[0]
    calculated_crc16 = crc16_func(complete_frame)

    if packet_crc16 != calculated_crc16:
        raise SMPBadCRC(f"Packet CRC {hex(packet_crc16)} != calculated CRC {hex(calculated_crc16)}")

    return complete_frame
